/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.gui;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;

import static java.lang.Math.cos;
import static java.lang.Math.sin;

public class Avatar implements ViewRenderListener {

    public static final double SPEED_LIMIT = 30;
    /**
     * Just in case we want to adjust global speed for some reason.
     */
    private static final double SPEED_MULTIPLIER = .02d;
    /**
     * Determines amount of friction user experiences every millisecond while moving around in space.
     */
    private static final double MILLISECOND_FRICTION = 1.005;
    /**
     * Avatar movement speed, relative to avatar itself. When avatar coordinates
     * are updated within the world, avatar orientation relative to the world is
     * taken into account.
     */
    private final Point3D movementVector = new Point3D();
    public double avatarAcceleration = 0.1;
    /**
     * Avatar location within the 3D world.
     */
    private Point3D location = new Point3D();
    /**
     * Avatar orientation on the X-Z plane. It changes when turning left or
     * right.
     */
    private double orientationXZ;
    /**
     * Avatar orientation on the Y-Z plane. It changes when looking up or down.
     */
    private double orientationYZ;

    public Avatar() {
    }

    public Avatar(final Avatar sourceView) {
        setLocation(new Point3D(sourceView.location));
        setAngleXZ(sourceView.getAngleXZ());
        setAngleYZ(sourceView.getAngleYZ());
    }

    public Avatar(final Point3D location) {
        setLocation(location);
    }

    public Avatar(final Point3D location, final float angleXZ,
                  final float angleYZ) {
        setLocation(location);
        setAngleXZ(angleXZ);
        setAngleYZ(angleYZ);
    }

    @Override
    public boolean beforeRender(final ViewPanel viewPanel, final int millisecondsSinceLastFrame) {

        final Point3D locationBeforeUpdate = new Point3D(location);
        translateAvatarLocationBasedOnMovementVector(millisecondsSinceLastFrame);
        applyFrictionToUserMovement(millisecondsSinceLastFrame);
        return isFrameRepaintNeeded(locationBeforeUpdate);
    }

    private boolean isFrameRepaintNeeded(Point3D locationBeforeUpdate) {
        final double distanceMoved = location.getDistanceTo(locationBeforeUpdate);
        return distanceMoved > 0.03;
    }

    public void enforceSpeedLimit() {
        final double currentSpeed = movementVector.getVectorLength();

        if (currentSpeed <= SPEED_LIMIT)
            return;

        movementVector.scaleDown(currentSpeed / SPEED_LIMIT);
    }

    public double getAngleXZ() {
        return orientationXZ;
    }

    public void setAngleXZ(final double angleXZ) {
        orientationXZ = angleXZ;
    }

    public double getAngleYZ() {
        return orientationYZ;
    }

    public void setAngleYZ(final double angleYZ) {
        orientationYZ = angleYZ;
    }

    public Point3D getLocation() {
        return location;
    }

    public void setLocation(final Point3D location) {
        this.location = location;
    }

    public Point3D getMovementVector() {
        return movementVector;
    }

    public double getMovementSpeed() {
        return movementVector.getVectorLength();
    }

    private void applyFrictionToUserMovement(int millisecondsPassedSinceLastFrame) {
        for (int i = 0; i < millisecondsPassedSinceLastFrame; i++)
            applyMillisecondFrictionToUserMovementVector();
    }

    private void applyMillisecondFrictionToUserMovementVector() {
        getMovementVector().x /= MILLISECOND_FRICTION;
        getMovementVector().y /= MILLISECOND_FRICTION;
        getMovementVector().z /= MILLISECOND_FRICTION;
    }

    /**
     * Translate coordinates based on avatar movement vector and avatar orientation in the world.
     *
     * @param millisecondsPassedSinceLastFrame We want avatar movement to be independent of framerate.
     *                                         Therefore we take frame rendering time into account when translating
     *                                         avatar between consecutive frames.
     */
    private void translateAvatarLocationBasedOnMovementVector(int millisecondsPassedSinceLastFrame) {
        location.x -= (float) sin(getAngleXZ())
                * getMovementVector().z * SPEED_MULTIPLIER
                * millisecondsPassedSinceLastFrame;
        location.z += (float) cos(getAngleXZ())
                * getMovementVector().z * SPEED_MULTIPLIER
                * millisecondsPassedSinceLastFrame;

        location.x += (float) cos(getAngleXZ())
                * getMovementVector().x * SPEED_MULTIPLIER
                * millisecondsPassedSinceLastFrame;
        location.z += (float) sin(getAngleXZ())
                * getMovementVector().x * SPEED_MULTIPLIER
                * millisecondsPassedSinceLastFrame;

        location.y += getMovementVector().y * SPEED_MULTIPLIER
                * millisecondsPassedSinceLastFrame;
    }
}
