/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.gui;

import eu.svjatoslav.sixth.e3d.gui.humaninput.MouseEvent;
import eu.svjatoslav.sixth.e3d.gui.humaninput.MouseInteractionController;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.WritableRaster;

public class RenderingContext {

    public static final int bufferedImageType = BufferedImage.TYPE_4BYTE_ABGR;
    public final Graphics2D graphics;
    public final byte[] pixels;
    public final int width;
    public final int height;
    public final int xCenter;
    public final int yCenter;
    public final double zoom;
    final BufferedImage bufferedImage;
    public int frameNumber = 0;

    /**
     * UI component that mouse is currently hovering over.
     */
    private MouseInteractionController currentMouseOverComponent;

    public void prepareForNewFrameRendering(){
        mouseEvent = null;
        objectUnderMouse = null;
    }

    /**
     * Mouse click event that needs to be processed.
     */
    private  MouseEvent mouseEvent;

    public void setMouseEvent(MouseEvent mouseEvent) {
        this.mouseEvent = mouseEvent;
    }

    public MouseEvent getMouseEvent() {
        return mouseEvent;
    }

    /**
     * Item that user clicked on.
     */
    private MouseInteractionController objectUnderMouse;

    /**
     * Called when given object was detected under mouse cursor, while processing {@link #mouseEvent}.
     * Because objects are rendered back to front. The last method caller will set the top-most object, if
     * there are multiple objects under mouse cursor.
     */
    public void setObjectUnderMouse(MouseInteractionController objectUnderMouse) {
        this.objectUnderMouse = objectUnderMouse;
    }

    public RenderingContext(final int width, final int height) {
        this.width = width;
        this.height = height;

        this.xCenter = width / 2;
        this.yCenter = height / 2;

        this.zoom = width / 3d;

        bufferedImage = new BufferedImage(width, height, bufferedImageType);

        final WritableRaster raster = bufferedImage.getRaster();
        final DataBufferByte dbi = (DataBufferByte) raster.getDataBuffer();
        pixels = dbi.getData();

        graphics = (Graphics2D) bufferedImage.getGraphics();
        graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        graphics.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    }

    /**
     * @return <code>true</code> if view repaint is needed.
     */
    public boolean handleDetectedComponentMouseEvents() {
        if (objectUnderMouse != null) {
            if (mouseEvent.button == 0) {
                // mouse over
                if (currentMouseOverComponent == null) {
                    currentMouseOverComponent = objectUnderMouse;
                    return currentMouseOverComponent.mouseEntered();
                } else if (currentMouseOverComponent != objectUnderMouse) {
                    boolean viewRepaintNeeded = currentMouseOverComponent.mouseExited();
                    currentMouseOverComponent = objectUnderMouse;
                    return viewRepaintNeeded | currentMouseOverComponent.mouseEntered();
                }
            } else {
                // mouse click
                return objectUnderMouse.mouseClicked();
            }
        } else if (currentMouseOverComponent != null) {
            boolean viewRepaintNeeded = currentMouseOverComponent.mouseExited();
            currentMouseOverComponent = null;
            return viewRepaintNeeded;
        }
        return false;
    }

}
