/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.gui.textEditorComponent;

import java.util.ArrayList;
import java.util.List;

public class TextLine {

    private List<Character> chars = new ArrayList<>();

    public TextLine() {
    }

    public TextLine(final List<Character> value) {
        chars = value;
        pack();
    }

    public TextLine(final String value) {
        setValue(value);
    }

    public void addIdent(final int amount) {
        if (isEmpty())
            return;

        for (int i = 0; i < amount; i++)
            chars.add(0, new Character(' '));
    }

    public String copySubString(final int from, final int until) {
        final StringBuilder result = new StringBuilder();

        ensureLength(until);

        for (int i = from; i < until; i++)
            result.append(chars.remove(from).value);

        pack();
        return result.toString();
    }

    public void cutFromBeginning(int charactersToCut) {

        if (charactersToCut > chars.size())
            charactersToCut = chars.size();

        if (charactersToCut == 0)
            return;

        chars = chars.subList(charactersToCut, chars.size());
    }

    public String cutSubString(final int from, final int until) {
        final StringBuilder result = new StringBuilder();

        final List<Character> reminder = new ArrayList<>();

        ensureLength(until);

        for (int i = 0; i < chars.size(); i++)
            if ((i >= from) && (i < until))
                result.append(chars.get(i).value);
            else
                reminder.add(chars.get(i));

        chars = reminder;

        pack();
        return result.toString();
    }

    public void cutUntilEnd(final int col) {
        if (col >= chars.size())
            return;

        chars = chars.subList(0, col);
    }

    private void ensureLength(final int length) {
        while (chars.size() < length)
            chars.add(new Character(' '));
    }

    public char getCharForLocation(final int col) {

        if (col >= chars.size())
            return ' ';

        return chars.get(col).value;
    }

    public List<Character> getChars() {
        return chars;
    }

    public int getIdent() {
        if (isEmpty())
            return 0;

        for (int i = 0; i < chars.size(); i++)
            if (chars.get(i).hasValue())
                return i;

        throw new RuntimeException("This code shall never execute");
    }

    public int getLength() {
        return chars.size();
    }

    public TextLine getSubLine(final int from, final int until) {
        final List<Character> result = new ArrayList<>();

        for (int i = from; i < until; i++) {
            if (i >= chars.size())
                break;
            result.add(chars.get(i));
        }

        return new TextLine(result);
    }

    public String getSubString(final int from, final int until) {
        final StringBuilder result = new StringBuilder();

        for (int i = from; i < until; i++)
            result.append(getCharForLocation(i));

        return result.toString();
    }

    public void insertCharacter(final int col, final char value) {
        ensureLength(col);
        chars.add(col, new Character(value));
        pack();
    }

    public void insertString(final int col, final String value) {
        ensureLength(col);
        int i = 0;
        for (final char c : value.toCharArray()) {
            chars.add(col + i, new Character(c));
            i++;
        }
        pack();
    }

    public void insertTextLine(final int col, final TextLine textLine) {
        ensureLength(col);
        int i = 0;
        for (final Character c : textLine.getChars()) {
            chars.add(col + i, c);
            i++;
        }
        pack();
    }

    public boolean isEmpty() {
        return chars.isEmpty();
    }

    private void pack() {
        int newLength = 0;

        for (int i = chars.size() - 1; i >= 0; i--)
            if (chars.get(i).hasValue()) {
                newLength = i + 1;
                break;
            }

        if (newLength == chars.size())
            return;

        chars = chars.subList(0, newLength);
    }

    public void removeCharacter(final int col) {
        if (col >= chars.size())
            return;

        chars.remove(col);
    }

    public void setValue(final String string) {
        chars.clear();
        for (final char c : string.toCharArray())
            chars.add(new Character(c));

        pack();
    }

    @Override
    public String toString() {
        final StringBuilder buffer = new StringBuilder();

        for (final Character character : chars)
            buffer.append(character.value);

        return buffer.toString();
    }

}
