/*
 * Sixth 3D engine. Copyright ©2012-2019, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.solidpolygon;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.gui.humaninput.MouseInteractionController;
import eu.svjatoslav.sixth.e3d.renderer.raster.Color;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;

import static eu.svjatoslav.sixth.e3d.geometry.Polygon.pointWithinPolygon;

public class SolidPolygon extends AbstractCoordinateShape {

    private final static LineInterpolator polygonBoundary1 = new LineInterpolator();
    private final static LineInterpolator polygonBoundary2 = new LineInterpolator();
    private final static LineInterpolator polygonBoundary3 = new LineInterpolator();
    private Color color;

    public SolidPolygon(final Point3D point1, final Point3D point2,
                        final Point3D point3, final Color color) {
        super(point1, point2, point3);
        this.color = color;
    }

    public static void drawHorizontalLine(final LineInterpolator line1,
                                          final LineInterpolator line2, final int y,
                                          final RenderingContext renderBuffer, final Color color) {

        int x1 = line1.getX(y);
        int x2 = line2.getX(y);

        if (x1 > x2) {
            final int tmp = x1;
            x1 = x2;
            x2 = tmp;
        }

        if (x1 < 0)
            x1 = 0;

        if (x2 >= renderBuffer.width)
            x2 = renderBuffer.width - 1;

        final int width = x2 - x1;

        int offset = ((y * renderBuffer.width) + x1) * 4;
        final byte[] offSreenBufferBytes = renderBuffer.pixels;

        final int polygonAlpha = color.a;
        final int b = color.b;
        final int g = color.g;
        final int r = color.r;

        if (polygonAlpha == 255)
            for (int i = 0; i < width; i++) {
                offSreenBufferBytes[offset] = (byte) 255;
                offset++;
                offSreenBufferBytes[offset] = (byte) b;
                offset++;
                offSreenBufferBytes[offset] = (byte) g;
                offset++;
                offSreenBufferBytes[offset] = (byte) r;
                offset++;
            }
        else {
            final int backgroundAlpha = 255 - polygonAlpha;

            final int blueWithAlpha = b * polygonAlpha;
            final int greenWithAlpha = g * polygonAlpha;
            final int redWithAlpha = r * polygonAlpha;

            for (int i = 0; i < width; i++) {
                offSreenBufferBytes[offset] = (byte) 255;
                offset++;
                offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + blueWithAlpha) / 256);
                offset++;
                offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + greenWithAlpha) / 256);
                offset++;
                offSreenBufferBytes[offset] = (byte) ((((offSreenBufferBytes[offset] & 0xff) * backgroundAlpha) + redWithAlpha) / 256);
                offset++;
            }

        }

    }

    public static void drawPolygon(final RenderingContext context,
                                   final Point2D onScreenPoint1, final Point2D onScreenPoint2,
                                   final Point2D onScreenPoint3,
                                   final MouseInteractionController mouseInteractionController,
                                   final Color color) {

        onScreenPoint1.roundToInteger();
        onScreenPoint2.roundToInteger();
        onScreenPoint3.roundToInteger();

        if (mouseInteractionController != null)
            if (context.getMouseEvent() != null)
                if (pointWithinPolygon(context.getMouseEvent().coordinate,
                        onScreenPoint1, onScreenPoint2, onScreenPoint3))
                    context.setObjectUnderMouse(mouseInteractionController);

        if (color.isTransparent())
            return;

        // find top-most point
        int yTop = (int) onScreenPoint1.y;

        if (onScreenPoint2.y < yTop)
            yTop = (int) onScreenPoint2.y;

        if (onScreenPoint3.y < yTop)
            yTop = (int) onScreenPoint3.y;

        if (yTop < 0)
            yTop = 0;

        // find bottom-most point
        int yBottom = (int) onScreenPoint1.y;

        if (onScreenPoint2.y > yBottom)
            yBottom = (int) onScreenPoint2.y;

        if (onScreenPoint3.y > yBottom)
            yBottom = (int) onScreenPoint3.y;

        if (yBottom >= context.height)
            yBottom = context.height - 1;

        // paint
        polygonBoundary1.setPoints(onScreenPoint1, onScreenPoint2);
        polygonBoundary2.setPoints(onScreenPoint1, onScreenPoint3);
        polygonBoundary3.setPoints(onScreenPoint2, onScreenPoint3);

        final LineInterpolator[] is = new LineInterpolator[3];
        is[0] = polygonBoundary1;
        is[1] = polygonBoundary2;
        is[2] = polygonBoundary3;

        java.util.Arrays.sort(is);

        for (int y = yTop; y < yBottom; y++)
            if (is[0].containsY(y)) {

                if (is[1].containsY(y))
                    drawHorizontalLine(is[0], is[1], y, context, color);
                else if (is[2].containsY(y))
                    drawHorizontalLine(is[0], is[2], y, context, color);
            } else if (is[1].containsY(y))
                if (is[2].containsY(y))
                    drawHorizontalLine(is[1], is[2], y, context, color);
    }

    public Color getColor() {
        return color;
    }

    public void setColor(final Color color) {
        this.color = color;
    }

    @Override
    public void paint(final RenderingContext renderBuffer) {

        final Point2D onScreenPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D onScreenPoint2 = coordinates[1].onScreenCoordinate;
        final Point2D onScreenPoint3 = coordinates[2].onScreenCoordinate;

        drawPolygon(renderBuffer, onScreenPoint1, onScreenPoint2,
                onScreenPoint3, mouseInteractionController, color);

    }

}
